/*
 *   $Id: device-linux.c,v 1.11 2004/10/25 20:38:38 psavola Exp $
 *
 *   Authors:
 *    Lars Fenneberg		<lf@elemental.net>	 
 *
 *   This software is Copyright 1996,1997 by the above mentioned author(s), 
 *   All Rights Reserved.
 *
 *   The license which is distributed with this software in the file COPYRIGHT
 *   applies to this software. If your distribution is missing this file, you
 *   may request it from <lutchann@litech.org>.
 *
 *    Greg Daley Sept 2004:   Added support for SLLAO NCE installs 
 */

#include <config.h>
#include <includes.h>
#include <radvd.h>
#include <defaults.h>
#include <pathnames.h>		/* for PATH_PROC_NET_IF_INET6 */

#ifndef NUD_VALID
#define NUD_VALID (NUD_REACHABLE|NUD_STALE|NUD_DELAY|NUD_PROBE)
#endif				/* NUD_VALID */

/*  rtnetlink socket includes */
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>

#ifndef IPV6_ADDR_LINKLOCAL
#define IPV6_ADDR_LINKLOCAL   0x0020U
#endif

static int new_nd_rtnetlink(int sock, int ifindex, unsigned char *dst,
			    unsigned char *lladdr, int lll, int state,
			    int replace);
static int process_ndmsg(struct ndmsg *ndm, int nll, unsigned char *dst,
			 unsigned char *lladdr, int lll);
static int check_nd_rtnetlink(int sock, int ifindex, unsigned char *dst,
			      unsigned char *lladdr, int lll);
static void clean_nl_inbuff(int sock);
static int get_nlsequence(void);
static int make_ndmsg(struct nlmsghdr *nlm, unsigned char *dst,
		      unsigned char *lladdr, int lll, int state,
		      int ifindex);

/*
 * this function gets the hardware type and address of an interface,
 * determines the link layer token length and checks it against
 * the defined prefixes
 */
int
setup_deviceinfo(int sock, struct Interface *iface)
{
	struct ifreq	ifr;
	struct AdvPrefix *prefix;
	char zero[HWADDR_MAX];
	
	strncpy(ifr.ifr_name, iface->Name, IFNAMSIZ-1);
	ifr.ifr_name[IFNAMSIZ-1] = '\0';

	if (ioctl(sock, SIOCGIFHWADDR, &ifr) < 0)
	{
		flog(LOG_ERR, "ioctl(SIOCGIFHWADDR) failed for %s: %s",
			iface->Name, strerror(errno));
		return (-1);
	}

	dlog(LOG_DEBUG, 3, "hardware type for %s is %d", iface->Name,
		ifr.ifr_hwaddr.sa_family); 

	switch(ifr.ifr_hwaddr.sa_family)
        {
	case ARPHRD_ETHER:
		iface->if_hwaddr_len = 48;
		iface->if_prefix_len = 64;
		iface->if_maxmtu = 1500;
		break;
#ifdef ARPHRD_FDDI
	case ARPHRD_FDDI:
		iface->if_hwaddr_len = 48;
		iface->if_prefix_len = 64;
		iface->if_maxmtu = 4352;
		break;
#endif /* ARPHDR_FDDI */
#ifdef ARPHRD_ARCNET
	case ARPHRD_ARCNET:
		iface->if_hwaddr_len = 8;
		iface->if_prefix_len = -1;
		iface->if_maxmtu = -1;
		break;
#endif /* ARPHDR_ARCNET */
	default:
		iface->if_hwaddr_len = -1;
		iface->if_prefix_len = -1;
		iface->if_maxmtu = -1;
		break;
	}

	dlog(LOG_DEBUG, 3, "link layer token length for %s is %d", iface->Name,
		iface->if_hwaddr_len);

	dlog(LOG_DEBUG, 3, "prefix length for %s is %d", iface->Name,
		iface->if_prefix_len);

	if (iface->if_hwaddr_len != -1) {
		memcpy(iface->if_hwaddr, ifr.ifr_hwaddr.sa_data, (iface->if_hwaddr_len + 7) >> 3);

		memset(zero, 0, (iface->if_hwaddr_len + 7) >> 3);
		if (!memcmp(iface->if_hwaddr, zero, (iface->if_hwaddr_len + 7) >> 3))
			flog(LOG_WARNING, "WARNING, MAC address on %s is all zero!",
				iface->Name);
	}

	prefix = iface->AdvPrefixList;
	while (prefix)
	{
		if ((iface->if_prefix_len != -1) &&
		   (iface->if_prefix_len != prefix->PrefixLen))
		{
			flog(LOG_WARNING, "prefix length should be %d for %s",
				iface->if_prefix_len, iface->Name);
 		}
 			
 		prefix = prefix->next;
	}
                
	return (0);
}

/*
 * this function extracts the link local address and interface index
 * from PATH_PROC_NET_IF_INET6
 */
int setup_linklocal_addr(int sock, struct Interface *iface)
{
	FILE *fp;
	char str_addr[40];
	unsigned int plen, scope, dad_status, if_idx;
	char devname[IFNAMSIZ];

	if ((fp = fopen(PATH_PROC_NET_IF_INET6, "r")) == NULL)
	{
		flog(LOG_ERR, "can't open %s: %s", PATH_PROC_NET_IF_INET6,
			strerror(errno));
		return (-1);	
	}
	
	while (fscanf(fp, "%32s %x %02x %02x %02x %15s\n",
		      str_addr, &if_idx, &plen, &scope, &dad_status,
		      devname) != EOF)
	{
		if (scope == IPV6_ADDR_LINKLOCAL &&
		    strcmp(devname, iface->Name) == 0)
		{
			struct in6_addr addr;
			unsigned int ap;
			int i;
			
			for (i=0; i<16; i++)
			{
				sscanf(str_addr + i * 2, "%02x", &ap);
				addr.s6_addr[i] = (unsigned char)ap;
			}
			memcpy(&iface->if_addr, &addr, sizeof(addr));

			iface->if_index = if_idx;
			fclose(fp);
			return 0;
		}
	}

	flog(LOG_ERR, "no linklocal address configured for %s", iface->Name);
	fclose(fp);
	return (-1);
}

int setup_allrouters_membership(int sock, struct Interface *iface)
{
	struct ipv6_mreq mreq;                  
	
	memset(&mreq, 0, sizeof(mreq));                  
	mreq.ipv6mr_interface = iface->if_index;
	
	/* ipv6-allrouters: ff02::2 */
	mreq.ipv6mr_multiaddr.s6_addr32[0] = htonl(0xFF020000);                                          
	mreq.ipv6mr_multiaddr.s6_addr32[3] = htonl(0x2);     

	if (setsockopt(sock, SOL_IPV6, IPV6_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
	{
		flog(LOG_ERR, "can't join ipv6-allrouters on %s", iface->Name);
		return (-1);
	}

	return (0);
}

int setup_routertorouter_membership(int sock, struct Interface *iface)
{
	struct ipv6_mreq mreq;                  
	
	memset(&mreq, 0, sizeof(mreq));                  
	mreq.ipv6mr_interface = iface->if_index;
	
	/* ipv6-allrouters: ff02::12 */
	mreq.ipv6mr_multiaddr.s6_addr32[0] = htonl(0xFF020000);                                          
	mreq.ipv6mr_multiaddr.s6_addr32[3] = htonl(0x12);     

	if (setsockopt(sock, SOL_IPV6, IPV6_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
	{
		flog(LOG_ERR, "can't join ipv6-routertorouters on %s", iface->Name);
		return (-1);
	}

	return (0);
}

int check_allrouters_membership(int sock, struct Interface *iface)
{
	#define ALL_ROUTERS_MCAST "ff020000000000000000000000000002"
	
	FILE *fp;
	unsigned int if_idx, allrouters_ok=0;
	char addr[32+1];
	int ret=0;

	if ((fp = fopen(PATH_PROC_NET_IGMP6, "r")) == NULL)
	{
		flog(LOG_ERR, "can't open %s: %s", PATH_PROC_NET_IGMP6,
			strerror(errno));
		return (-1);	
	}
	
	while ( (ret=fscanf(fp, "%4u %*s %32[0-9A-Fa-f] %*x %*x %*x\n", &if_idx, addr)) != EOF) {
		if (ret == 2) {
			if (iface->if_index == if_idx) {
				if (strncmp(addr, ALL_ROUTERS_MCAST, sizeof(addr)) == 0)
					allrouters_ok = 1;
			}
		}
	}

	fclose(fp);

	if (!allrouters_ok) {
		flog(LOG_WARNING, "resetting ipv6-allrouters membership on %s", iface->Name);
		setup_allrouters_membership(sock, iface);
	}	

	return(0);
}		

int check_routertorouter_membership(int sock, struct Interface *iface)
{
	#define R2ROUTER_MCAST "ff020000000000000000000000000012"
	
	FILE *fp;
	unsigned int if_idx, r2router_ok=0;
	char addr[32+1];
	int ret=0;

	if ((fp = fopen(PATH_PROC_NET_IGMP6, "r")) == NULL)
	{
		flog(LOG_ERR, "can't open %s: %s", PATH_PROC_NET_IGMP6,
			strerror(errno));
		return (-1);	
	}
	
	while ( (ret=fscanf(fp, "%4u %*s %32[0-9A-Fa-f] %*x %*x %*x\n", &if_idx, addr)) != EOF) {
		if (ret == 2) {
			if (iface->if_index == if_idx) {
				if (strncmp(addr, R2ROUTER_MCAST, sizeof(addr)) == 0)
					r2router_ok = 1;
			}
		}
	}

	fclose(fp);

	if (!r2router_ok) {
		flog(LOG_WARNING, "resetting ipv6-r2router membership on %s", iface->Name);
		setup_routertorouter_membership(sock, iface);
	}	

	return(0);
}		

int
get_v4addr(const char *ifn, unsigned int *dst)
{
	struct ifreq	ifr;
	struct sockaddr_in *addr;
	int fd;

	if( ( fd = socket(AF_INET,SOCK_DGRAM,0) ) < 0 )
	{
		flog(LOG_ERR, "create socket for IPv4 ioctl failed for %s: %s",
			ifn, strerror(errno));
		return (-1);
	}
	
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifn, IFNAMSIZ-1);
	ifr.ifr_name[IFNAMSIZ-1] = '\0';
	ifr.ifr_addr.sa_family = AF_INET;
	
	if (ioctl(fd, SIOCGIFADDR, &ifr) < 0)
	{
		flog(LOG_ERR, "ioctl(SIOCGIFADDR) failed for %s: %s",
			ifn, strerror(errno));
		close( fd );
		return (-1);
	}

	addr = (struct sockaddr_in *)(&ifr.ifr_addr);

	dlog(LOG_DEBUG, 3, "IPv4 address for %s is %s", ifn,
		inet_ntoa( addr->sin_addr ) ); 

	*dst = addr->sin_addr.s_addr;

	close( fd );

	return 0;
}

int
bind_rtnlsock(void)
{
	int rtnetlink_socket;
	struct sockaddr_nl local;

	memset(&local, 0, sizeof(local));
	local.nl_family = AF_NETLINK;

	if ((rtnetlink_socket =
	     socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE))
	    < 0) {
		flog(LOG_ERR, "Unable to get a netlink socket: %s",
		     strerror(errno));
		return -1;
	}

	if (bind
	    (rtnetlink_socket, (struct sockaddr *) &local,
	     sizeof(local)) < 0) {
		flog(LOG_ERR, "Unable to bind a netlink socket: %s",
		     strerror(errno));
		return -1;
	}
	return rtnetlink_socket;
}

/**************************************************************************/




static int
new_nd_rtnetlink(int sock, int ifindex, unsigned char *dst,
			    unsigned char *lladdr, int lll, int state,
			    int replace)
{
	struct {
		struct nlmsghdr nh;
		struct ndmsg nd;
		char attrbuf[360];
	} req;

	struct ndmsg *ndm;
	struct nlmsghdr *nlm;
	int nll;

	memset(&req, 0, sizeof(req));

	nlm = (struct nlmsghdr *) &req;

	nlm->nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE;

	nlm->nlmsg_flags |= (replace ? NLM_F_REPLACE : NLM_F_EXCL);
	nlm->nlmsg_seq = get_nlsequence();
	nlm->nlmsg_type = RTM_NEWNEIGH;
	ndm = NLMSG_DATA(nlm);
	nlm->nlmsg_len = make_ndmsg(nlm, dst, lladdr, lll, state, ifindex);
	nll = nlm->nlmsg_len;

	if (send(sock, &req, req.nh.nlmsg_len, 0) < 0) {
		flog(LOG_ERR, "unable to send netlink message: %s",
		     strerror(errno));
		return -1;
	}
	/*system("/sbin/ip -f inet6 neigh show");*/
	return 0;
}


static int
process_ndmsg(struct ndmsg *ndm, int nll, unsigned char *dst,
			 unsigned char *lladdr, int lll)
{
	int ret = 0;
	unsigned int mlll = 0;
	unsigned char *m_lladdr = NULL;
	unsigned char *m_dst = NULL;
	struct rtattr *rta;

	rta = NDA_RTA(ndm);

	dlog(LOG_DEBUG, 5, "NCE process i(%d) t(%d)", ndm->ndm_ifindex,
	     ndm->ndm_type);

	while (RTA_OK(rta, nll)) {
		switch (rta->rta_type) {
		case NDA_DST:
			dlog(LOG_DEBUG, 4,
			     "ent: %x:%x:%x:%x:%x:%x:%x:%x",
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[0]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[1]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[2]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[3]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[4]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[5]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[6]),
			     ntohs(((struct in6_addr *)
				    RTA_DATA(rta))->s6_addr16[7]));

			dlog(LOG_DEBUG, 4,
			     "dst: %x:%x:%x:%x:%x:%x:%x:%x",
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[0]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[1]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[2]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[3]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[4]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[5]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[6]),
			     ntohs(((struct in6_addr *)
				    dst)->s6_addr16[7]));

			if (!memcmp(RTA_DATA(rta), dst,
				    sizeof(struct in6_addr))) {
				m_dst = RTA_DATA(rta);
			}
			break;
		case NDA_LLADDR:
			m_lladdr = RTA_DATA(rta);
			mlll = RTA_PAYLOAD(rta);
			break;
		default:
			break;
		}
		rta = RTA_NEXT(rta, nll);
	}
	if (m_dst) {
		/* we have the correct entry */
		ret = -2;
		if (ndm->ndm_state & NUD_VALID) {
			if (m_lladdr && (mlll == lll) && !memcmp(lladdr,
								 m_lladdr,
								 lll *
								 sizeof
								 (unsigned
								  char))) {
				/* entry matches, update with same state */
				flog(LOG_ERR, "state:%x "
				     "%02x:%02x:%02x:%02x:%02x:%02x",
				     ndm->ndm_state,
				     m_lladdr[0], m_lladdr[1],
				     m_lladdr[2], m_lladdr[3],
				     m_lladdr[4], m_lladdr[5]);
				ret = ndm->ndm_state;
			} else {	/* conflicting NCE ignores */
				ret = -1;
			}
		}
	}
	return ret;
}



static int
check_nd_rtnetlink(int sock, int ifindex, unsigned char *dst,
			      unsigned char *lladdr, int lll)
{
	int recvsize;
	struct {
		struct nlmsghdr nh;
		struct ndmsg nd;
		char attrbuf[3000];
	} req;
	unsigned char *m_lladdr = NULL;

	struct ndmsg *ndm;
	struct nlmsghdr *nlm;
	int nll;
	int ret = 0;

	memset(&req, 0, sizeof(struct nlmsghdr) + sizeof(struct ndmsg));

	nlm = (struct nlmsghdr *) &req;
	nlm->nlmsg_flags = NLM_F_REQUEST | NLM_F_MATCH;
	nlm->nlmsg_seq = get_nlsequence();
	nlm->nlmsg_type = RTM_GETNEIGH;
	ndm = NLMSG_DATA(nlm);
	ndm->ndm_family = AF_INET6;
	ndm->ndm_ifindex = ifindex;
	ndm->ndm_type = RTN_UNICAST;

	nlm->nlmsg_len = make_ndmsg(nlm, dst, lladdr, lll, 0, ifindex);
	nll = nlm->nlmsg_len;

	if (send(sock, &req, nll, 0) < 0) {
		flog(LOG_ERR, "unable to send netlink message: %s",
		     strerror(errno));
		return -1;
	}

	while ((recvsize = recv(sock, &req, sizeof(req), 0)) > 0) {
		nlm = (struct nlmsghdr *) &req;
		flog(LOG_WARNING,
		     "received rtnetlink info of length %d %x", recvsize,
		     nlm->nlmsg_flags);


		while (NLMSG_OK(nlm, recvsize)) {

			switch (nlm->nlmsg_type) {
			case RTM_NEWNEIGH:
				m_lladdr = NULL;
				ndm = NLMSG_DATA(nlm);

				nll = nlm->nlmsg_len;

				if ((ret =
				    process_ndmsg(ndm, nll, dst, lladdr,
						  lll))) {
					if (ret == -2)
						ret = 0;
					clean_nl_inbuff(sock);
					return ret;
				}
				break;
			case NLMSG_ERROR:
				flog(LOG_ERR,
				     "netlink error message received: %s",
				     strerror(-((struct nlmsgerr *)
						NLMSG_DATA(&req))->error));
				return 0;
			case NLMSG_DONE:
				flog(LOG_WARNING,
				     "netlink done message received");
				return ret;
			default:
				break;
			}
			nlm = NLMSG_NEXT(nlm, recvsize);
		}
	}
	if (recvsize < 0) {
		flog(LOG_ERR, "unable to receive netlink message: %s",
		     strerror(errno));
		return -1;
	}

	return 0;
}


static void
clean_nl_inbuff(int sock)
{
	char buff[1500];
	int recvsize;
	struct nlmsghdr *nlm;

	while ((recvsize = recv(sock, &buff, sizeof(buff), 0)) > 0) {
		nlm = (struct nlmsghdr *) &buff;
		while (NLMSG_OK(nlm, recvsize)) {
			switch (nlm->nlmsg_type) {
			case NLMSG_DONE:
				return;
				break;
			default:
				break;
			}
			nlm = NLMSG_NEXT(nlm, recvsize);
		}
	}
	if (recvsize < 0) {
		flog(LOG_ERR, "unable to receive netlink message: %s",
		     strerror(errno));
		return;
	}
}

static int
get_nlsequence(void)
{
	static int seq = 500;
	return ++seq;
}


static int
make_ndmsg(struct nlmsghdr *nlm, unsigned char *dst,
		      unsigned char *lladdr, int lll, int state,
		      int ifindex)
{
	struct rtattr *rta;
	unsigned long *rtap;
	int i;
	struct ndmsg *ndm;
	int nll;

	ndm = NLMSG_DATA(nlm);
	ndm->ndm_family = AF_INET6;
	ndm->ndm_type = RTN_UNICAST;
	ndm->ndm_ifindex = ifindex;
	ndm->ndm_state = state;

	nlm->nlmsg_len = NLMSG_LENGTH(sizeof(struct ndmsg));
	nll = nlm->nlmsg_len;

	rta = NDA_RTA(ndm);

	rta->rta_type = NDA_DST;
	rta->rta_len = RTA_LENGTH(sizeof(struct in6_addr));
	rtap = (unsigned long *) RTA_DATA(rta);

	for (i = 0; i < 4; i++) {
		/*rtap[i] =  ntohl(((struct in6_addr *)dst)->s6_addr32[i]); */
		rtap[i] = ((struct in6_addr *) dst)->s6_addr32[i];
	}
	/*memcpy(RTA_DATA(rta),dst,sizeof(struct in6_addr)); */

	nlm->nlmsg_len =
	    NLMSG_ALIGN(nlm->nlmsg_len) + NLMSG_ALIGN(rta->rta_len);
	nll = nlm->nlmsg_len;

	rta = RTA_NEXT(rta, nll);

	rta->rta_type = NDA_LLADDR;
	rta->rta_len = RTA_LENGTH(lll * sizeof(unsigned char));
	memcpy(RTA_DATA(rta), lladdr, lll);

	nlm->nlmsg_len =
	    NLMSG_ALIGN(nlm->nlmsg_len) + NLMSG_ALIGN(rta->rta_len);
	nll = nlm->nlmsg_len;

	return nll;
}


void
install_nc_entry(int sock, struct Interface *iface, struct in6_addr *addr,
		 unsigned char *hwaddr, int hwl, int replace)
{
	int state;

	dlog(LOG_DEBUG, 4, "installing NCE on %s   replace: %d, len %d",
	     iface->Name, replace, hwl);


	state =
	    check_nd_rtnetlink(sock, iface->if_index,
			       (unsigned char *) addr, hwaddr, hwl);

	if (state > 0) {
		dlog(LOG_DEBUG, 4, "entry match\n");
		if (new_nd_rtnetlink
		    (sock, iface->if_index, (unsigned char *) addr, hwaddr,
		     hwl, state, replace) < 0) {

			flog(LOG_ERR, "unable to update neighbour cache");
			return;
		}
	} else {
		dlog(LOG_DEBUG, 4,
		     "no entry or mismatch SLLAO --> STALE\n");
		if (new_nd_rtnetlink
		    (sock, iface->if_index, (unsigned char *) addr, hwaddr,
		     hwl, NUD_STALE, replace) < 0) {
			flog(LOG_ERR, "unable to %s neighbour cache entry",
			     (replace ? "replace" : "create"));
			return;
		}
	}
}
