/*
 *   $Id: process.c,v 1.8 2004/10/26 05:30:34 psavola Exp $
 *
 *   Authors:
 *    Pedro Roque		<roque@di.fc.ul.pt>
 *    Lars Fenneberg		<lf@elemental.net>	 
 *
 *   This software is Copyright 1996,1997 by the above mentioned author(s), 
 *   All Rights Reserved.
 *
 *   The license which is distributed with this software in the file COPYRIGHT
 *   applies to this software. If your distribution is missing this file, you
 *   may request it from <lutchann@litech.org>.
 *
 */

#include <config.h>
#include <includes.h>
#include <radvd.h>
#include <r2router.h>
#include <detfra.h>
#include <securend.h>


extern int rtnl_sock;

static void process_rs(int sock, struct Interface *iface, unsigned char *msg,
			int len, struct sockaddr_in6 *addr);
static unsigned char *process_rs_options(struct Interface *, unsigned char *msg,
			int len, struct sockaddr_in6 *, uint32_t *flags,
			struct timeval *tv);

static void process_ra(struct Interface *, unsigned char *msg, int len,
			struct sockaddr_in6 *);
static int  addr_match(struct in6_addr *a1, struct in6_addr *a2,
			int prefixlen);
static int rs_has_srclladdr(unsigned char *, int);
static void process_r2r(int sock, struct Interface *iface, unsigned char *msg,
			int len, struct sockaddr_in6 *addr);

void
process(int sock, struct Interface *ifacel, unsigned char *msg, int len, 
	struct sockaddr_in6 *addr, struct in6_pktinfo *pkt_info, int hoplimit)
{
	struct Interface *iface;
	struct icmp6_hdr *icmph;
	char addr_str[INET6_ADDRSTRLEN];

	if ( ! pkt_info )
	{
		flog(LOG_WARNING, "received packet with no pkt_info!" );
		return;
	}

	/*
	 * can this happen?
	 */

	if (len < sizeof(struct icmp6_hdr))
	{
		flog(LOG_WARNING, "received icmpv6 packet with invalid length: %d",
			len);
		return;
	}

	icmph = (struct icmp6_hdr *) msg;

	if (icmph->icmp6_type != ND_ROUTER_SOLICIT &&
	    icmph->icmp6_type != ND_ROUTER_TO_ROUTER &&
	    icmph->icmp6_type != ND_ROUTER_ADVERT)
	{
		/*
		 *	We just want to listen to RSs and RAs
		 */
		
		flog(LOG_ERR, "icmpv6 filter failed");
		return;
	}

	if (icmph->icmp6_type == ND_ROUTER_ADVERT)
	{
		if (len < sizeof(struct nd_router_advert)) {
			flog(LOG_WARNING, "received icmpv6 RA packet with invalid length: %d",
				len);
			return;
		}

		if (!IN6_IS_ADDR_LINKLOCAL(&addr->sin6_addr)) {
			flog(LOG_WARNING, "received icmpv6 RA packet with non-linklocal source address");
			return;
		}
	}			
	
	if (icmph->icmp6_type == ND_ROUTER_SOLICIT)
	{
		if (len < sizeof(struct nd_router_solicit)) {
			flog(LOG_WARNING, "received icmpv6 RS packet with invalid length: %d",
				len);
			return;
		}
		
		if (IN6_IS_ADDR_UNSPECIFIED(&addr->sin6_addr) &&
		    rs_has_srclladdr(msg, len))
		{
			flog(LOG_WARNING, "received icmpv6 RS packet with unspecified source address and there is a lladdr option"); 
			return;
		}
				
	}			

        if( icmph->icmp6_type == ND_ROUTER_TO_ROUTER)
        {
		if (len < sizeof(struct nd_router_to_router)) {
			flog(LOG_WARNING, "received icmpv6 R2R packet with invalid length: %d",
				len);
			return;
		}

		if (!IN6_IS_ADDR_LINKLOCAL(&addr->sin6_addr)) {
			flog(LOG_WARNING, "received icmpv6 R2R packet with non-linklocal source address");
			return;
		}
	        dlog(LOG_DEBUG, 4, "R2R received %d", pkt_info->ipi6_ifindex);

        }
        else /* if */
	if (icmph->icmp6_code != 0)
	{
		flog(LOG_WARNING, "received icmpv6 RS/RA packet with invalid code: %d",
			icmph->icmp6_code);
		return;
	}
	
	dlog(LOG_DEBUG, 4, "if_index %d", pkt_info->ipi6_ifindex);

	/* get iface by received if_index */

	for (iface = ifacel; iface; iface=iface->next)
	{
		if (iface->if_index == pkt_info->ipi6_ifindex)
		{
			break;
		}
	}

	if (iface == NULL)
	{
		dlog(LOG_DEBUG, 2, "received packet from unknown interface: %d",
			pkt_info->ipi6_ifindex);
		return;
	}
	
	if (hoplimit != 255)
	{
		print_addr(&addr->sin6_addr, addr_str);
		flog(LOG_WARNING, "received RS or RA with invalid hoplimit %d from %s",
			hoplimit, addr_str);
		return;
	}
	
	if (!iface->AdvSendAdvert)
	{
		dlog(LOG_DEBUG, 2, "AdvSendAdvert is off for %s", iface->Name);
		return;
	}

	dlog(LOG_DEBUG, 4, "found Interface: %s", iface->Name);

	if (icmph->icmp6_type == ND_ROUTER_SOLICIT)
	{
                process_rs(sock, iface, msg, len, addr);
	}
	else if (icmph->icmp6_type == ND_ROUTER_ADVERT)
	{
		process_ra(iface, msg, len, addr);
	}
        else if (icmph->icmp6_type == ND_ROUTER_TO_ROUTER)
	{
		process_r2r(sock, iface, msg, len, addr);
	}
}

static void
process_rs(int sock, struct Interface *iface, unsigned char *msg,
		int len, struct sockaddr_in6 *addr)
{
	struct in6_addr unspec = { { { 0 } } };
	struct in6_addr bitdiff;
	int delay=0;
        int delayset=0;
	double next, resdelay;
	struct timeval tv;
	int tokens =-1;
        int do_multicast=1;
        struct dtimer_list *dtm_tmp=NULL;
        unsigned char *outopts=NULL;
	int calcnr=0, probslot=0;
        uint32_t flags=0;
	int rank;

	gettimeofday(&tv, NULL);

	dlog(LOG_DEBUG, 5, "PROCESS ROUTER SOLICIT ON %s", iface->Name);


	outopts=process_rs_options(iface, msg, len, addr, &flags, &tv); 

	if(iface->ProbabilisticFastRA||iface->DetermineFastRA||
		iface->UnnegotiatedFastRA)
		tokens = token_bucket_update(iface, &tv);

	if(!delayset && iface->UnnegotiatedFastRA && tokens &&
		 memcmp(&addr->sin6_addr, &unspec, sizeof(struct in6_addr))){
			
		rank=check_before(iface->RouterList,iface->CurrFastRARouters,
					(uint32_t *)&iface->FastRAOwnHashBits,
					&addr->sin6_addr);
		iface->NumUnicastFastRAs--;
		delay=rank * iface->FastRASeparation;
		dlog(LOG_DEBUG, 3, "Unneg FastRA delay for %s %d r(%d)",
				iface->Name, delay, rank);
                delayset=1;
	}


	if(!delayset && iface->ProbabilisticFastRA && tokens)
	{
		dlog(LOG_DEBUG, 5, "ProbFastRA on %s? %d f(%8.8x) ", 
			iface->Name, iface->ProbabilisticFastRA, flags);

		iface->NumUnicastFastRAs--;
		calcnr= PROB_NUM_ROUTERS_ESTIMATE( 
		iface->ProbabilisticNumRouters );

		/* generates a range 0..(N^2-1) */
		probslot=rand()%(calcnr * calcnr);

		if(probslot < calcnr)
		{
			delay = 0;
		}
		else
		{
			delay = (((probslot - calcnr)/(calcnr - 1)) + 1) *
						iface->ProbFastRASlot;
		}
		dlog(LOG_DEBUG, 3, "ProbFastRA delay for %s: %d (N: %d)",
				iface->Name, delay, calcnr);
                delayset=1;
	}

	if(!delayset && iface->DetermineFastRA && 
			( iface->r2rState & R2R_STATES_READY) && tokens)
	{
		iface->NumUnicastFastRAs--;
		delay=iface->dfraSelf.FixedDelay * 1.0 ;
		dlog(LOG_DEBUG, 3, "DetFastRA delay for %s: %d (%d)", 
				iface->Name, delay, iface->NumUnicastFastRAs);
		delayset=1;
        }
        if(!delayset){
		delay = (int) (MAX_RA_DELAY_TIME*rand()/(RAND_MAX+1.0));
  		dlog(LOG_DEBUG, 3, "random mdelay for %s: %d", iface->Name, delay);
	}
   	
	next=tv_diff(&(iface->tm.expires), &tv);

	/* residual delay for multicast */
	resdelay = MAX2((((double)delay)/1000.0),
			MAX2(0.0, (iface->MinDelayBetweenRAs) - 
				tv_diff(&tv,&iface->last_multicast)));
 
	dlog(LOG_DEBUG, 4, "next:%10.10f resdelay: %10.10f (%f) [%f]",next,
		resdelay,MAX2(0.0, (iface->MinDelayBetweenRAs) -
			tv_diff(&tv,&iface->last_multicast)),
		iface->MinDelayBetweenRAs - tv_diff(&tv,&iface->last_multicast) );
        /* GD: not sure here: on no more fastras need multicast! */
        if(!tokens){
		dlog(LOG_DEBUG, 4, "Run out of FastRA tokens on %s, Mcast",
					 iface->Name);
	}
	else if((!IN6_IS_ADDR_UNSPECIFIED(&addr->sin6_addr) && 
		(resdelay > ((double)delay)/1000.0))|| iface->UnicastOnly ||
		delayset){
		dlog(LOG_DEBUG, 4, "too soon for mcast on %s", iface->Name);
		do_multicast=0; 
	} 


	/* MCast rescheduling */

	if(do_multicast){
		dlog(LOG_DEBUG, 5, "multicast response on %s",iface->Name);
		/* Reschedule multicast if nearer */
		if((resdelay + 0.001) < next){
			dlog(LOG_DEBUG, 3, "Reschedule multicast to %f on %s",
						resdelay,iface->Name);
			next=resdelay;
			clear_timer(&iface->tm);
			set_timer(&iface->tm, next);
		}
		if(outopts)
			dlog(LOG_DEBUG, 4, "check nonce AWAITS %p %d %d %d %d",
				outopts, iface->responseAwaitLen,
				outopts[1]<<3, MAX_ResponseOptionLen,
				((iface->responseAwaitLen + (outopts[1]<<3)) <=
					MAX_ResponseOptionLen));

		/* Add any nonces to output list */
	        if(iface->NonceResponse && outopts && (outopts[1] < 3) && 
			((iface->responseAwaitLen + (outopts[1]<<3))
						<= MAX_ResponseOptionLen)){

			memcpy(iface->responseAwaitBuf +iface->responseAwaitLen,
				outopts, outopts[1]<<3); 
			iface->responseAwaitLen +=  outopts[1]<<3; 
		}
		return;
	} 
 
  		/*
  		 *	unicast reply
  		 */
  
	if(delay == 0){ /* No need to take a schedule slot */
		dlog(LOG_DEBUG, 5, "Zero Delay response %s",iface->Name);
		send_ra_flags(sock, iface, &addr->sin6_addr, outopts, flags);
		return;
        }
	dlog(LOG_DEBUG, 3, "delay on  %s %d",iface->Name, delay);
	if((dtm_tmp = remove_from_list(&iface->dyntimers[FREE_Dynamic_Timers],
			NULL)) ==  NULL){
		dlog(LOG_DEBUG, 3, "No entries on free list %s",iface->Name);
		return;
  	}
	if(iface->NonceResponse && outopts && (outopts[1]<3)){
		memcpy(dtm_tmp->options, outopts, outopts[1]<<3); 
		dtm_tmp->flags=flags;
	}
	memcpy(&dtm_tmp->addr, &addr->sin6_addr, sizeof (struct in6_addr));
  
	insert_into_list(&iface->dyntimers[USED_Dynamic_Timers], dtm_tmp);
  		
	/* start: remove from production */
	log_list(&iface->dyntimers[FREE_Dynamic_Timers], "FREE p");
	log_list(&iface->dyntimers[USED_Dynamic_Timers], "USED p");
	/* end: remove from production */
      
	next = ((double)delay)/1000.0; 

	set_timer(dtm_tmp->tm, next);
	return;
}

/*
 * check router advertisements according to RFC 2461, 6.2.7
 */
static void
process_ra(struct Interface *iface, unsigned char *msg, int len, 
	struct sockaddr_in6 *addr)
{
	struct nd_router_advert *radvert;
        unsigned int advint=0;
	char addr_str[INET6_ADDRSTRLEN];
	int matched=0;
	int new_landmark=0;
        int new_overload=0;
	struct timeval tv,now;
        int num; 

	uint8_t *opt_str;

	print_addr(&addr->sin6_addr, addr_str);

	if(!memcmp(&addr->sin6_addr, &iface->if_addr, sizeof(struct in6_addr))){
		dlog(LOG_DEBUG, 5, "own RA from %s on %s", addr_str,
                                iface->Name);
		return;
	}

	gettimeofday(&now, NULL);


	dlog(LOG_DEBUG, 1, "RECEIVE ROUTER ADV from %s on %s", addr_str,
				iface->Name);


	radvert = (struct nd_router_advert *) msg;

	if ((radvert->nd_ra_curhoplimit && iface->AdvCurHopLimit) && 
	   (radvert->nd_ra_curhoplimit != iface->AdvCurHopLimit))
	{
		flog(LOG_WARNING, "our AdvCurHopLimit on %s doesn't agree with %s",
			iface->Name, addr_str);
	}

	if ((radvert->nd_ra_flags_reserved & ND_RA_FLAG_MANAGED) && !iface->AdvManagedFlag)
	{
		flog(LOG_WARNING, "our AdvManagedFlag on %s doesn't agree with %s",
			iface->Name, addr_str);
	}
	
	if ((radvert->nd_ra_flags_reserved & ND_RA_FLAG_OTHER) && !iface->AdvOtherConfigFlag)
	{
		flog(LOG_WARNING, "our AdvOtherConfigFlag on %s doesn't agree with %s",
			iface->Name, addr_str);
	}

	/* note: we don't check the default router preference here, because they're likely different */

	if ((radvert->nd_ra_reachable && iface->AdvReachableTime) &&
	   (ntohl(radvert->nd_ra_reachable) != iface->AdvReachableTime))
	{
		flog(LOG_WARNING, "our AdvReachableTime on %s doesn't agree with %s",
			iface->Name, addr_str);
	}
	
	if ((radvert->nd_ra_retransmit && iface->AdvRetransTimer) &&
	   (ntohl(radvert->nd_ra_retransmit) != iface->AdvRetransTimer))
	{
		flog(LOG_WARNING, "our AdvRetransTimer on %s doesn't agree with %s",
			iface->Name, addr_str);
	}

	dlog(LOG_DEBUG, 1, "About to check length %d %d", len, 
				 sizeof(struct nd_router_advert));
	len -= sizeof(struct nd_router_advert);

	if (len == 0)
		return;
		
	opt_str = (uint8_t *)(msg + sizeof(struct nd_router_advert));
		
	while (len > 0)
	{
		int optlen;
		struct nd_opt_prefix_info *pinfo;
		struct nd_opt_mtu *mtu;
		struct AdvPrefix *prefix;
		char prefix_str[INET6_ADDRSTRLEN];
		uint32_t preferred, valid;

		if (len < 2)
		{
			flog(LOG_ERR, "trailing garbage in RA on %s from %s", 
				iface->Name, addr_str);
			break;
		}
		
		optlen = (opt_str[1] << 3);

		if (optlen == 0) 
		{
			flog(LOG_ERR, "zero length option in RA on %s from %s",
				iface->Name, addr_str);
			break;
		} 
		else if (optlen > len)
		{
			flog(LOG_ERR, "option length greater than total"
				" length in RA on %s from %s",
				iface->Name, addr_str);
			break;
		} 		

		switch (*opt_str)
		{
		case ND_OPT_MTU:
			mtu = (struct nd_opt_mtu *)opt_str;

			if (iface->AdvLinkMTU && (ntohl(mtu->nd_opt_mtu_mtu) != iface->AdvLinkMTU))
			{
				flog(LOG_WARNING, "our AdvLinkMTU on %s doesn't agree with %s",
					iface->Name, addr_str);
			}
			break;
		case ND_OPT_PREFIX_INFORMATION:
			pinfo = (struct nd_opt_prefix_info *) opt_str;
			preferred = ntohl(pinfo->nd_opt_pi_preferred_time);
			valid = ntohl(pinfo->nd_opt_pi_valid_time);
			matched=0;
			
			print_addr(&pinfo->nd_opt_pi_prefix, prefix_str);
			dlog(LOG_DEBUG,5,"received PIO: %s/%d",
				prefix_str, pinfo->nd_opt_pi_prefix_len);

			prefix = iface->AdvPrefixList;
			while (prefix)
			{
				if (prefix->enabled &&
				    (prefix->PrefixLen == pinfo->nd_opt_pi_prefix_len) &&
				    addr_match(&prefix->Prefix, &pinfo->nd_opt_pi_prefix,
				    	 prefix->PrefixLen))
				{
					print_addr(&prefix->Prefix, prefix_str);
					matched=1;

					if (valid != prefix->AdvValidLifetime)
					{
						flog(LOG_WARNING, "our AdvValidLifetime on"
						 " %s for %s doesn't agree with %s",
						 iface->Name,
						 prefix_str,
						 addr_str
						 );
					}
					if (preferred != prefix->AdvPreferredLifetime)
					{
						flog(LOG_WARNING, "our AdvPreferredLifetime on"
						 " %s for %s doesn't agree with %s",
						 iface->Name,
						 prefix_str,
						 addr_str
						 );
					}
				}
				dlog(LOG_DEBUG,5,"PIO match: %d",matched);
				prefix = prefix->next;
			}			
			if(!matched && iface->PrefixLandmarking){
				struct in6_addr prefaddr;
				/* add prefix to landmark list */
				prefix_mask(&prefaddr,
					&pinfo->nd_opt_pi_prefix,
					pinfo->nd_opt_pi_prefix_len);
				print_addr(&prefaddr, prefix_str);

				if((num = add_rtrprefix(iface->PrefixLandList,
					iface->MaxPrefixLandmarks,	
					iface->CurrPrefixLandmarks,	
					&prefaddr,
					preferred,0))){
					
					new_landmark=1;
					iface->CurrPrefixLandmarks=num;
					dlog(LOG_DEBUG,5,"Landmark %s %d",
						prefix_str,
						iface->CurrPrefixLandmarks);
				}		
				else{
					/* overload */
					new_overload=1;
					dlog(LOG_DEBUG,5,"Overload Landmark %s/%d",
						prefix_str,
						pinfo->
						nd_opt_pi_prefix_len);
				}
			}
			break;
		case ND_OPT_ROUTE_INFORMATION:
			/* not checked: these will very likely vary a lot */
			break;
		case ND_OPT_LINK_ID:
		case ND_OPT_NONCE:
		case ND_OPT_SOURCE_LINKADDR:
			/* not checked */
			break;
		case ND_OPT_TARGET_LINKADDR:
		case ND_OPT_REDIRECTED_HEADER:
			flog(LOG_ERR, "invalid option %d in RA on %s from %s",
				(int)*opt_str, iface->Name, addr_str);
			break;
		/* Mobile IPv6 extensions */
		case ND_OPT_RTR_ADV_INTERVAL:
			advint=ntohl(((struct nd_opt_adv_interval *)opt_str)->nd_opt_adv_interval_ival);
			dlog(LOG_DEBUG, 5, "Advint %d in RA on %s from %s",
                                advint, iface->Name, addr_str);
		case ND_OPT_HOME_AGENT_INFO:
			/* not checked */
			break;
		default:
			dlog(LOG_DEBUG, 1, "unknown option %d in RA on %s from %s",
				(int)*opt_str, iface->Name, addr_str);
			break;
		}
		
		len -= optlen;
		opt_str += optlen;
	}
	/* GD XXX update with processed Advertisement Interval */
	if(iface->ProbabilisticFastRA && (!iface->ProbFastRAProutersOnly 
		||(radvert->nd_ra_flags_reserved & ND_RA_FLAG_PROB_FASTRA) ))
		add_to_prtrlist(iface,&addr->sin6_addr, advint?
					MIN2(advint, 300000): 300000);

	/* GD XXX update with processed Advertisement Interval */
	if(iface->UnnegotiatedFastRA && (radvert->nd_ra_flags_reserved &
						 ND_RA_FLAG_FASTRA)){
		uint32_t life;	
		
		life = (advint/1000 ? MIN2(advint/1000, 
                                ntohs(radvert->nd_ra_router_lifetime)):
                           	ntohs(radvert->nd_ra_router_lifetime));	
		
		if((num =  add_rtrprefix(iface->RouterList,
                                        	iface->MaxFastRARouters,
                                        	iface->CurrFastRARouters,
                                        	&addr->sin6_addr, life,1))){
	
			iface->CurrFastRARouters=num;
		}
	}
			

	if(iface->PrefixLandmarking){
		
		tv.tv_usec = tv.tv_sec=0;

		if(advint && (new_overload || new_landmark)){
			struct timeval ai_timeval;

                	ai_timeval.tv_usec = (advint % 1000) * 1000;
                	ai_timeval.tv_sec = advint / 1000;
			timeradd(&now, &ai_timeval, &tv);
		}
		if(new_overload){
			if(tv_diff(&tv, &iface->LandOverloadExpiry) >0){
				dlog(LOG_DEBUG, 1, "new overload timer: %d.%d",
					tv.tv_sec, tv.tv_usec);
				memcpy(&iface->LandOverloadExpiry,
					&tv, sizeof (struct timeval));
			}
		}	
		/*if(new_landmark){
			update_landtimers(iface->PrefixLandList,
						iface->CurrPrefixLandmarks,&tv);
		}*/
	}
        

}

static int 
addr_match(struct in6_addr *a1, struct in6_addr *a2, int prefixlen)
{
	int pdw;
	int pbi;

	pdw = prefixlen >> 0x05;  /* num of whole uint32_t in prefix */
	pbi = prefixlen &  0x1f;  /* num of bits in incomplete uint32_t in prefix */

	if (pdw) 
	{
		if (memcmp(a1, a2, pdw << 2))
			return 0;
	}

	if (pbi) 
	{
		uint32_t w1, w2;
		uint32_t mask;

		w1 = *((uint32_t *)a1 + pdw);
		w2 = *((uint32_t *)a2 + pdw);

		mask = htonl(((uint32_t) 0xffffffff) << (0x20 - pbi));

		if ((w1 ^ w2) & mask)
			return 0;
	}

	return 1;
}

static int
rs_has_srclladdr(unsigned char *msg, int len)
{
	uint8_t *opt_str;
	int	has_lladdr = 0;

	len -= sizeof(struct nd_router_solicit);

	if (len == 0)
	{		/* no option */
		return 0;
	}

	opt_str = (uint8_t *)(msg + sizeof(struct nd_router_solicit));

	while (len > 0)
	{
		int optlen;

		if (len < 2)
		{
			flog(LOG_WARNING, "trailing garbage in RS");
			break;
		}

		optlen = (opt_str[1] << 3);

		if (optlen == 0)
		{
			flog(LOG_WARNING, "zero length option in RS");
			break;
		}
		else if (optlen > len)
		{
			flog(LOG_WARNING, "option length greater than total length in RS");
			break;
		}

		if (*opt_str == ND_OPT_SOURCE_LINKADDR)
		{
			has_lladdr = 1;
		}

		len -= optlen;
		opt_str += optlen;
	}

	return has_lladdr;
}




static unsigned char *process_rs_options(struct Interface *iface, 
	unsigned char *msg, int len, struct sockaddr_in6 *addr, uint32_t *flags,
	struct timeval *tv)
{
	uint8_t *opt_str;
	int i;
	unsigned char *copy=NULL;
	unsigned char hw_addr[16];
	struct AdvPrefix *prefix=NULL;
	struct nd_opt_gpl *lm_opt=NULL;
	struct in6_addr tmp_addr;
 
	len -= sizeof(struct nd_router_solicit);
	opt_str = (uint8_t *)(msg + sizeof(struct nd_router_solicit));

	while (len > 0)
	{
		int optlen;

		if (len < 2)
		{
			flog(LOG_ERR, "trailing garbage in RS");
			break;
		}

		optlen = (opt_str[1] << 3);

		dlog(LOG_DEBUG,4, "option type %d, length %d", (int)*opt_str,
			optlen);

		if (optlen == 0)
		{
			flog(LOG_ERR, "zero length option in RA");
			break;
		}
		else if (optlen > len)
		{
			flog(LOG_ERR, "option length greater than total"
				" length in RA (type %d, optlen %d, len %d)",
				(int)*opt_str, optlen, len);
			break;
		}
		switch (*opt_str)
		{
                case ND_OPT_SOURCE_LINKADDR:
                        dlog(LOG_DEBUG, 4, "RecvSourceLLAddress: %d l:%d",
                              iface->RecvSourceLLAddress, optlen);

                        if(iface->RecvSourceLLAddress && (optlen <= 16)){
                              for (i = 2; i < optlen; i++)
                              {
                                              hw_addr[i-2] = opt_str[i];
                              }
                                install_nc_entry(rtnl_sock,iface,
                                              &addr->sin6_addr, hw_addr, 6,1);
                        }
                        break;
                case ND_OPT_TENTATIVE_SLLA:
                        dlog(LOG_DEBUG, 4, "RecvTentativeLLAddress: ");

                        if(iface->RecvTentativeLLAddress && (optlen <= 16)){
                                      for (i = 2; i < optlen; i++)
                              {
                                              hw_addr[i-2] = opt_str[i];
                              }
                                install_nc_entry(rtnl_sock,iface,
                                              &addr->sin6_addr, hw_addr, 6,0);
                        }
                        break;
		case ND_OPT_NONCE:
			dlog(LOG_DEBUG, 4, "received Nonce option in RS on %s",
				iface->Name);
			copy=opt_str;
			break;
		case ND_OPT_LANDMARK:
			dlog(LOG_DEBUG, 4, "received Landmark option in RS on %s",
				iface->Name);
		 	lm_opt = (struct nd_opt_gpl *)opt_str;
			/* check if a known prefix */
			prefix = iface->AdvPrefixList;
			while (prefix)
			{
				if (prefix->enabled &&
					addr_match(&prefix->Prefix, 
						(struct in6_addr *)&lm_opt->nd_opt_gpl_prefix, 64))
				{
					*flags |= ND_RA_FLAG_LANDMARK_SEEN;  
					break;
				}
				prefix=prefix->next;
			}
			if(*flags & ND_RA_FLAG_LANDMARK_SEEN)
				break; 

			/* store in an ipv6 address for comparison */
			memset(&tmp_addr, 0, sizeof (struct in6_addr));
			tmp_addr.s6_addr32[0]=
				lm_opt->nd_opt_gpl_prefix.p6_addr32[0];
			tmp_addr.s6_addr32[1]=
				lm_opt->nd_opt_gpl_prefix.p6_addr32[1];

			if(check_landmark(iface->PrefixLandList,
                                        	iface->CurrPrefixLandmarks,
						&tmp_addr,tv)){
				*flags |= ND_RA_FLAG_LANDMARK_SEEN;
			}
			
			
			/* if not, indicate invalid landmark, unless overload*/
			if(! (*flags & ND_RA_FLAG_LANDMARK_SEEN )){ 
				
				dlog(LOG_DEBUG,5, "Unseen: Now: %d.%d OE:%d.%d (td %f)",
					tv->tv_sec,tv->tv_usec,
					iface->LandOverloadExpiry.tv_sec,
					iface->LandOverloadExpiry.tv_usec,
					tv_diff(tv, &iface->LandOverloadExpiry));
				if (tv_diff(tv, &iface->LandOverloadExpiry) >0)
				{
					/* timer expired, know not seen */
					*flags |= ND_RA_FLAG_LANDMARK_INVALID;  
				}	
			}
			break;
		default:
			dlog(LOG_DEBUG, 1, "unknown option %d in RS",
				(int)*opt_str);
			break;
		}

		len -= optlen;
		opt_str += optlen;
 
	}
	return copy;
}
 

static void
process_r2r(int sock, struct Interface *iface, unsigned char *msg, int len, 
	struct sockaddr_in6 *addr)
{
	struct nd_router_to_router *r2r;
	char addr_str[INET6_ADDRSTRLEN];
        struct nd_opt_detfra   *detfrao=NULL;
        struct nd_opt_linkid   *lid=NULL;
        struct AdvInterval     *ival=NULL;
	uint8_t *opt_str;

	print_addr(&addr->sin6_addr, addr_str);

        if(!memcmp(&addr->sin6_addr,&iface->if_addr,sizeof(struct in6_addr))){
		dlog(LOG_DEBUG, 4, "Receive R2R from ourself on %s (%s)",
				iface->Name, addr_str);
                return;
        }

	r2r = (struct nd_router_to_router *) msg;

	if (!iface->RouterToRouter)
	{ /* draft-daley-dna-det-fastra-00.txt 11.3: silent discard */
		return;
	}

	dlog(LOG_DEBUG, 4, "Receive R2R from %s on %s", addr_str, iface->Name);

        /* Other Validity Checks */

	len -= sizeof(struct nd_router_to_router);

	if (len == 0)
		return;
		
	opt_str = (uint8_t *)(msg + sizeof(struct nd_router_to_router));
		
	while (len > 0)
	{
		int optlen;

		if (len < 2)
		{
			flog(LOG_ERR, "trailing garbage in R2R on %s from %s", 
				iface->Name, addr_str);
			break;
		}
		
		optlen = (opt_str[1] << 3);

		if (optlen == 0) 
		{
			flog(LOG_ERR, "zero length option in R2R on %s from %s",
				iface->Name, addr_str);
			break;
		} 
		else if (optlen > len)
		{
			flog(LOG_ERR, "option length greater than total"
				" length in R2R on %s from %s",
				iface->Name, addr_str);
			break;
		} 		

		switch (*opt_str)
		{
		/* deterministic fast RA options */
		case ND_OPT_DET_FAST_RA:
                        detfrao=(struct nd_opt_detfra *)opt_str;
			break;
		case ND_OPT_LINK_ID:
                        lid=(struct nd_opt_linkid *)opt_str;
			break;
		case ND_OPT_SOURCE_LINKADDR:
			dlog(LOG_DEBUG, 4, "SLLAO in R2R on %s from %s",
				iface->Name, addr_str);
			/* not checked */
			break;
		case ND_OPT_NONCE:
			/* not checked */
			break;
		/* Mobile IPv6 extensions */
		case ND_OPT_RTR_ADV_INTERVAL:
			dlog(LOG_DEBUG, 4, "Router advertisment interval in R2R on %s from %s",
				iface->Name, addr_str);
			/* not checked */
                        ival = (struct AdvInterval *)opt_str; 
			break;
		default:
			dlog(LOG_DEBUG, 1, "unknown option %d in R2R on %s from %s",
				(int)*opt_str, iface->Name, addr_str);
			break;
		}
		
		len -= optlen;
		opt_str += optlen;
	}
        if(detfrao && iface->DetermineFastRA){
               if(ival){
                     r2r_handledetfrao(iface, detfrao, addr,
						ntohl(ival->adv_ival));
               }
               else{
                     r2r_handledetfrao(iface, detfrao, addr, (int)((
			iface->RouterToRouterStatusInt + 
			MAX_RTR_STATUS_DELAY) * 1000) );
               }
        }
	if(lid && iface->UseLinkID){
		dlog(LOG_DEBUG, 4, "LinkID in R2R on %s from %s %04x%04x%04x %d"
			, iface->Name, addr_str,
                      	ntohs(lid->nd_opt_linkid_id.link_id16[0]),
                       	ntohs(lid->nd_opt_linkid_id.link_id16[1]),
                       	ntohs(lid->nd_opt_linkid_id.link_id16[2]),
			linkid_compare(&iface->ProspectiveLinkID, 
					&lid->nd_opt_linkid_id));
		r2r_handlelinkid(iface, lid);
	}
          /* This responding Status MUST be sent after a random delay of */
          /*  0 to MAX_RTR_STATUS_DELAY                                  */
          /* Additionally, multicast Status Messages MUST                */
          /* NOT be sent more regularly than once every                  */
          /* MIN_DELAY_BETWEEN_RTR_STATUS on average.                    */
          /* Responses to Status-Request messages MUST be sent(GD:if trusted)*/
          /* send responses to mc group if MIN_DELAY_BETWEEN_RTR_STATUS */
          /* since last (sec 5.2 draft-daley-dna-det-fastra-00.txt)     */
	if (r2r->nd_r2r_type == ND_ROUTER_TO_ROUTER_STATUS_REQ)
	{
                int delay;
                struct timeval tv;

                delay =(int)(MAX_RTR_STATUS_DELAY*rand()/(RAND_MAX+1.0));

                dlog(LOG_DEBUG, 3, "random rtr mdelay for %s: %d", iface->Name,
				 delay);
                mdelay(delay);
 
                gettimeofday(&tv, NULL);
                if (tv_diff(&tv, &iface->last_rtrmc) >=
					 MIN_DELAY_BETWEEN_RTR_STATUS)
                {
                        send_r2r(sock, iface,NULL ,1);
                }
                else{
                        send_r2r(sock, iface,&addr->sin6_addr ,1);
                }
	}
        
}

